from skimage.registration import phase_cross_correlation # change in scipy 0.19
# image = data.camera()
from skimage import transform
import numpy as np

__DEBUG__ = False

def apply_translation(img, y, x):
    """
    Applies translation to an image.

    Args:
        img (numpy.ndarray): The input image.
        y (float): Translation along the y-axis.
        x (float): Translation along the x-axis.

    Returns:
        numpy.ndarray: The translated image.
    """
    afine_tf = transform.AffineTransform(translation=(x, y))
    translated = transform.warp(img, inverse_map=afine_tf, order=0, preserve_range=True)
    return translated

def pre_register_images(orig_t0, orig_t1, apply_to_orig_t0=False):
    """
    Performs pre-registration between two images.

    Args:
        orig_t0 (numpy.ndarray): The first image.
        orig_t1 (numpy.ndarray): The second image.
        apply_to_orig_t0 (bool, optional): Whether to apply the translation to orig_t0. Defaults to False.

    Returns:
        tuple or numpy.ndarray: If apply_to_orig_t0 is False, returns the global translations (gloabl_t0, gloabl_t1).
        If apply_to_orig_t0 is True, returns the translated orig_t0 image.

    Notes:
        - The function uses phase cross-correlation to estimate the global translation between the two images.
        - The global translation is then applied to either orig_t0 or both orig_t0 and orig_t1, depending on the
          value of apply_to_orig_t0.
    """
    global_translation, error, diffphase = phase_cross_correlation(orig_t0, orig_t1)
    gloabl_t0 = -global_translation[0]
    gloabl_t1 = -global_translation[1]
    if not apply_to_orig_t0:
        return gloabl_t0, gloabl_t1
    else:
        return apply_translation(orig_t0, -gloabl_t0, -gloabl_t1)

def get_pyramidal_registration(orig_t0, orig_t1, depth=2, threshold_translation=20):
    """
    Performs pyramidal registration between two images.

    Args:
        orig_t0 (numpy.ndarray): The first image.
        orig_t1 (numpy.ndarray): The second image.
        depth (int, optional): The depth of the pyramid. Defaults to 2.
        threshold_translation (float, optional): The threshold for ignoring large translations. Defaults to 20.

    Returns:
        numpy.ndarray: The translation matrix representing the translations for each pixel.

    Notes:
        - The function applies pre-registration between orig_t0 and orig_t1.
        - It then iteratively performs registration on blocks of decreasing size.
        - Translations are accumulated in the translation matrix.
    """

    translation_matrix = np.zeros((*orig_t0.shape, 2))

    gloabl_t0, gloabl_t1 = pre_register_images(orig_t0, orig_t1)

    translation_matrix[..., 0] += gloabl_t0
    translation_matrix[..., 1] += gloabl_t1

    height = orig_t0.shape[0]
    width = orig_t0.shape[1]

    block_size = 256

    for i in range(depth):
        try:
            tile_height = block_size // (i * 2)
            tile_width = block_size // (i * 2)
        except:
            tile_height = block_size
            tile_width = block_size

        for y in range(0, height, tile_height):
            for x in range(0, width, tile_width):
                afine_tf = transform.AffineTransform(
                    translation=(-translation_matrix[y, x, 1], -translation_matrix[y, x, 0]))
                translated = transform.warp(orig_t0, inverse_map=afine_tf, order=0, preserve_range=True)

                tile_t0 = translated[y:y + tile_height, x:x + tile_width]
                tile_t1 = orig_t1[y:y + tile_height, x:x + tile_width]

                shift, error, diffphase = phase_cross_correlation(tile_t0, tile_t1)
                t0 = -shift[0]
                t1 = -shift[1]

                if abs(t0) > threshold_translation or abs(t1) > threshold_translation:
                    continue

                translation_matrix[y:y + tile_height, x:x + tile_width, 0] += t0
                translation_matrix[y:y + tile_height, x:x + tile_width, 1] += t1

    return translation_matrix



# TODO --> do random crops that through reflect can create images of any given size --> smarter for training
# TODO --> try the pipes now

# MEGA TODO CHECK THAT THIS IS REALLY SMART to set EXTRAPOLATION_MASKS to 1 instead of 0 !!!

# this contains all the data augmentations
# and should make sure the same nb of random are called so that everything keeps being in sync even when piped --> TODO
from functools import partial
from datetime import datetime
import matplotlib.pyplot as plt
from skimage import filters
from skimage.util import random_noise

import epyseg.img
from epyseg.img import Img, mask_rows_or_columns, elastic_deform, \
    create_random_intensity_graded_perturbation, apply_2D_gradient, array_equal, array_different, pad_border_xy, \
    interpolation_free_rotation
import random
from scipy import ndimage
from skimage import transform
import numpy as np
from skimage import exposure
from epyseg.tools.logger import TA_logger # logging
# TODO add augmentation randomly swap channels --> can learn by itself which channel is containing epithelia
from epyseg.utils.commontools import execute_chained_functions


logger = TA_logger()

# TODO --> find trick to get it to work as I want !!! --> below need be tunable parameters
EXTRAPOLATION_MASKS = 1  # shall I set it to 1 or TO 0 --> think about it important in fact --> THIS MAY NEED BE CHANGED --> can maybe also get it from the kwargs --> TODO

# see how to set the range since it's global !!! is it reset upon a new call ????
# if so need add kwargs to all --> to get the default range
def blur(orig, is_mask, **kwargs):
    """
    Apply Gaussian blur to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The blurred image or the original mask.

    # Examples:
    #     >>> image = np.array([[1, 2, 3],
    #     ...                   [4, 5, 6],
    #     ...                   [7, 8, 9]])
    #     >>> result = blur(image, False, blur=1.5)
    #     >>> print(result)
    #     [[1.00132017 2.00099013 3.00066008]
    #      [4.00033004 5.         5.99966996]
    #      [6.99933992 7.99900987 8.99867983]]
    """
    # Generate a random Gaussian blur value within the specified range
    gaussian_blur = random.uniform(0, kwargs['blur'])

    if not is_mask:
        # Apply Gaussian blur to the input image
        if len(orig.shape) == 4:
            # Apply 2D Gaussian blur to each slice of the 3D image
            for n, slc in enumerate(orig):
                orig[n] = filters.gaussian(slc, gaussian_blur, preserve_range=True, mode='wrap')
            out = orig
        else:
            # Apply 2D Gaussian blur to the input image
            out = filters.gaussian(orig, gaussian_blur, preserve_range=True, mode='wrap')
    else:
        # Return the original mask without blurring
        out = orig

    return out


def graded_intensity_modification(orig, is_mask, **kwargs):
    """
    Modify the intensity of the input image or mask using a graded perturbation.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The modified image or the original mask.

    # Examples:
    #     >>> image = np.array([[1, 2, 3],
    #     ...                   [4, 5, 6],
    #     ...                   [7, 8, 9]])
    #     >>> result = graded_intensity_modification(image, False)
    #     >>> print(result)
    #     [[0 1 2]
    #      [3 4 5]
    #      [6 7 8]]
    """
    # Apply a graded perturbation to the input image or mask
    out = apply_2D_gradient(orig, create_random_intensity_graded_perturbation(orig, off_centered=True, mock=is_mask))

    return out



# TODO --> make use of that
# nb could do random crops too
# nb does crop work for images with channels ??? not so sure --> need check
def crop_augmentation(orig, is_mask, coords_dict, **kwargs):
    """
    Perform crop augmentation on the input image or mask based on the provided coordinates.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        coords_dict (dict): A dictionary containing the crop coordinates.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The cropped image or mask.

    # Examples:
    #     >>> image = np.array([[1, 2, 3],
    #     ...                   [4, 5, 6],
    #     ...                   [7, 8, 9]])
    #     >>> coords = {'x1': 0, 'x2': 2, 'y1': 0, 'y2': 2}
    #     >>> result = crop_augmentation(image, False, coords)
    #     >>> print(result)
    #     [[1 2]
    #      [4 5]]
    """
    # Extract the crop coordinates
    startx = coords_dict['x1']
    starty = coords_dict['y1']

    # Random crop if startx and starty are None
    if startx is None:
        try:
            width = coords_dict['w']
        except:
            width = coords_dict['width']
        min = orig.shape[-2] - width
        if min < 0:
            min = 0
        startx = random.randrange(start=0, stop=min, step=1)
    if starty is None:
        try:
            height = coords_dict['h']
        except:
            height = coords_dict['height']
        min = orig.shape[-3] - height
        if min < 0:
            min = 0
        starty = random.randrange(start=0, stop=min, step=1)

    # Determine the end coordinates
    if 'w' in coords_dict or 'width' in coords_dict:
        if 'w' in coords_dict:
            endx = startx + coords_dict['w']
        else:
            endx = startx + coords_dict['width']
    else:
        endx = coords_dict['x2']
    if 'h' in coords_dict or 'height' in coords_dict:
        if 'h' in coords_dict:
            endy = starty + coords_dict['h']
        else:
            endy = starty + coords_dict['height']
    else:
        endy = coords_dict['y2']

    # Adjust for negative coordinates
    if endx < 0:
        endx = orig.shape[-2] + endx
    if endy < 0:
        endy = orig.shape[-3] + endy

    # Swap coordinates if necessary
    if starty > endy:
        tmp = starty
        starty = endy
        endy = tmp
    if startx > endx:
        tmp = startx
        startx = endx
        endx = tmp

    # Crop the image or mask based on the coordinates
    if len(orig.shape) == 3:
        return orig[starty:endy, startx:endx]
    else:
        return orig[::, starty:endy, startx:endx]


def invert(orig, is_mask, **kwargs):
    """
    Invert the colors of the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The inverted image or mask.

    Examples:
        >>> image = np.array([[0, 1, 0],
        ...                   [1, 0, 1],
        ...                   [0, 1, 0]])
        >>> result = invert(image, False)
        >>> print(result)
        [[1 0 1]
         [0 1 0]
         [1 0 1]]
    """
    if not is_mask:
        inverted_image =  epyseg.img.invert(orig)
    else:
        inverted_image = orig
    return inverted_image


def rotate(orig, is_mask, **kwargs):
    """
    Rotate the input image or mask by a random angle.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The rotated image or mask.

    # Examples:
    #     >>> import math
    #     >>> image = np.array([[1, 2, 3],
    #     ...                   [4, 5, 6],
    #     ...                   [7, 8, 9]])
    #     >>> result = rotate(image, False,**{'rotate':0.2})
    #     >>> print(result)
    #     [[7 4 1]
    #      [8 5 2]
    #      [9 6 3]]
    """
    # Rotate by random angle between 0 and 360
    angle = random.randint(0, kwargs['rotate'] * 360)
    angle *= 1 if random.random() < 0.5 else -1
    order = random.choice([0, 1, 3])

    final_order = order
    if is_mask:
        final_order = EXTRAPOLATION_MASKS

    # Bug fix for 3D images
    # Rotating the image using ndimage.rotate with the specified angle, axes, reshape, and order
    rot_orig = ndimage.rotate(orig, angle, axes=(-3, -2), reshape=False, order=final_order)

    return rot_orig


def strong_elastic_with_zoom_and_translation(orig, is_mask, **kwargs):
    """
    Apply strong elastic deformation with zoom and translation to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The transformed image or mask.

    # Examples:
    #     >>> image = np.array([[1, 2, 3],
    #     ...                   [4, 5, 6],
    #     ...                   [7, 8, 9]])
    #     >>> result = strong_elastic_with_zoom_and_translation(image, False)
    #     >>> print(result)
    #     [[5 5 5]
    #      [5 5 5]
    #      [5 5 5]]
    """
    # Apply elastic deformation with zoom and translation
    return elastic_with_zoom_and_translation(orig, is_mask, strength=100, **kwargs)



def elastic_with_zoom_and_translation(orig, is_mask, strength=25, **kwargs):
    """
    Apply elastic deformation with zoom and translation to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        strength (int): The strength of the elastic deformation.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The transformed image or mask.


    """
    # Set numpy random seed based on the provided random seed
    np.random.seed(random.randint(0, 10000000))

    # Generate a random displacement matrix
    displacement = np.random.randn(2, 3, 3) * strength

    # Select the order for interpolation
    order = random.choice([0, 1])

    # Generate random rotation and zoom values
    rotation = random.randint(0, kwargs['rotate'] * 360)
    rotation *= 1 if random.random() < 0.5 else -1
    zoom = random.uniform(1. - kwargs['zoom'], 1. + kwargs['zoom'])

    # Adjust order for masks
    if is_mask:
        order = EXTRAPOLATION_MASKS

    # Apply elastic deformation with rotation and zoom
    if len(orig.shape) == 4:
        # Assume 'dhwc' format
        rot_orig = elastic_deform(orig, displacement, axis=(1, 2), order=order, rotation=rotation, zoom=zoom)
    else:
        # Assume 'hwc' format
        rot_orig = elastic_deform(orig, displacement, axis=(0, 1), order=order, rotation=rotation, zoom=zoom)

    return rot_orig


def strong_elastic(orig, is_mask, **kwargs):
    """
    Apply strong elastic deformation to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The transformed image or mask.

    """
    # Call elastic function with a strength of 100
    return elastic(orig, is_mask, strength=100, **kwargs)

def elastic(orig, is_mask, strength=25, **kwargs):
    """
    Apply elastic deformation to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        strength (float): The strength of the elastic deformation.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The transformed image or mask.

    """
    # Set the random seed for numpy based on the random seed to ensure reproducibility
    np.random.seed(random.randint(0, 10000000))

    # Generate a random deformation matrix based on the strength parameter
    displacement = np.random.randn(2, 3, 3) * strength

    # Select the order for interpolation
    order = random.choice([0, 1])

    # Adjust the order for masks
    if is_mask:
        order = EXTRAPOLATION_MASKS

    # Apply elastic deformation based on the dimensions of the input
    if len(orig.shape) == 4:
        # Assume 'dhwc'
        transformed_orig = elastic_deform(orig, displacement, axis=(1, 2), order=order)
    else:
        # Assume 'hwc'
        transformed_orig = elastic_deform(orig, displacement, axis=(0, 1), order=order)

    return transformed_orig


# nb will that create a bug when images don't have same width and height --> most likely yes but test it --> TODO
# in fact I should not allow that except if image has same width and height or I should use a crop of it --> can I do that ???
def rotate_interpolation_free(orig, is_mask, **kwargs):
    """
    Apply rotation to the input image or mask without using interpolation.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The rotated image or mask.

    # Examples:
    #     >>> image = np.array([[1, 2, 3],
    #     ...                   [4, 5, 6],
    #     ...                   [7, 8, 9]])
    #     >>> result = rotate_interpolation_free(image, False)
    #     >>> print(result)
    #     [[3 6 9]
    #      [2 5 8]
    #      [1 4 7]]
    """
    # Determine the rotation angle based on the dimensions of the input
    if orig.shape[-2] != orig.shape[-3]:
        angle = 180
    else:
        angle = random.choice([90, 180, 270])

    # Apply interpolation-free rotation using the provided angle
    rot_orig = interpolation_free_rotation(orig, angle=angle)

    return rot_orig


def translate(orig, is_mask, **kwargs):
    """
    Apply translation to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The translated image or mask.

    """
    # Generate random translation values
    trans_x = np.random.randint(-int(orig.shape[-2] * kwargs['translate']),
                                int(orig.shape[-2] * kwargs['translate']))
    trans_y = np.random.randint(-int(orig.shape[-3] * kwargs['translate']),
                                int(orig.shape[-3] * kwargs['translate']))

    # Choose the interpolation order randomly
    order = random.choice([0, 1, 3])

    # Create an affine transformation matrix for translation
    afine_tf = transform.AffineTransform(translation=(trans_y, trans_x))

    # Determine the final interpolation order based on the input type
    final_order = order
    if is_mask:
        final_order = EXTRAPOLATION_MASKS

    # Apply translation to the input image or mask
    if len(orig.shape) <= 3:
        zoomed_orig = transform.warp(orig, inverse_map=afine_tf, order=final_order, preserve_range=True,
                                     mode='reflect')
    else:
        zoomed_orig = np.zeros_like(orig, dtype=orig.dtype)
        for i, slice in enumerate(orig):
            zoomed_orig[i] = transform.warp(slice, inverse_map=afine_tf, order=final_order, preserve_range=True,
                                            mode='reflect')

    return zoomed_orig


def stretch(orig, is_mask,**kwargs):
    raise Exception('TODO --> needs a lot of reworking !!!')
    scale = random.uniform(self.stretch_range[0],
                           self.augmentation_types_and_values['stretch'])
    order = random.choice([0, 1, 3])
    orientation = random.choice([0, 1])
    if orientation == 0:
        afine_tf = transform.AffineTransform(scale=(scale, 1))
    else:
        afine_tf = transform.AffineTransform(scale=(1, scale))

    final_order = order
    if is_mask:
        final_order = EXTRAPOLATION_MASKS
        # TODO check that I'm not doing unnecessary stuff here but seems to work... maybe recheck dtype too
        if self.is_output_1px_wide:
            dilation = 0
            if scale >= 2 and scale <= 3:
                # do 1 dilat
                dilation = 1
            elif scale >= 3:
                # do 2 dilat
                dilation = 2
            if dilation != 0:
                s = ndimage.generate_binary_structure(2, 1)
                # Apply dilation to every channel then reinject
                for c in range(orig.shape[-1]):
                    dilated = orig[..., c]
                    if len(orig.shape) == 4:
                        for n, slc in enumerate(dilated):
                            for dil in range(dilation):
                                dilated[n] = ndimage.grey_dilation(slc, footprint=s)
                        orig[..., c] = dilated
                    else:
                        for dil in range(dilation):
                            dilated = ndimage.grey_dilation(dilated, footprint=s)
                        orig[..., c] = dilated

    if len(orig.shape) == 4:
        # handle 3D images
        for n, slc in enumerate(orig):
            orig[n] = transform.warp(slc, inverse_map=afine_tf, order=final_order, preserve_range=True,
                                     mode='reflect')
        stretched_orig = orig
    else:
        stretched_orig = transform.warp(orig, inverse_map=afine_tf, order=final_order, preserve_range=True,
                                        mode='reflect')  # nb should I do this per channel to avoid issues --> MAYBE

    # print('test3', stretched_orig.max(), stretched_orig.min())
    return stretched_orig

# CODER KEEP TIP: SCALE IS INVERTED (COMPARED TO MY OWN LOGIC SCALE 2 MEANS DIVIDE SIZE/DEZOOM BY 2 AND 0.5 MEANS INCREASE SIZE/ZOOM BY 2 --> # shall I use 1/scale
def zoom(orig, is_mask, **kwargs):
    """
    Apply zooming to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The zoomed image or mask.

    """
    # Generate a random scale factor for zooming
    scale = random.uniform(1. - kwargs['zoom'], 1. + kwargs['zoom'])

    # Choose the interpolation order randomly
    order = random.choice([0, 1, 3])

    # Create an affine transformation matrix for zooming
    afine_tf = transform.AffineTransform(scale=(scale, scale))

    # Determine the final interpolation order based on the input type
    final_order = order
    if is_mask:
        final_order = EXTRAPOLATION_MASKS

    # Apply zooming to the input image or mask
    if len(orig.shape) == 4:
        # Handle 3D images
        for n, slc in enumerate(orig):
            orig[n] = transform.warp(slc, inverse_map=afine_tf, order=final_order, preserve_range=True,
                                     mode='reflect')
        zoomed_orig = orig
    else:
        zoomed_orig = transform.warp(orig, inverse_map=afine_tf, order=final_order, preserve_range=True,
                                     mode='reflect')

    return zoomed_orig


def shear(orig, is_mask, **kwargs):
    """
    Apply shearing to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The sheared image or mask.

    """
    # Generate a random shear factor within the specified range
    shear = random.uniform(-kwargs['shear'], kwargs['shear'])

    # Choose the interpolation order randomly
    order = random.choice([0, 1, 3])

    # Create an affine transformation matrix for shearing
    afine_tf = transform.AffineTransform(shear=shear)

    # Determine the final interpolation order based on the input type
    final_order = order
    if is_mask:
        final_order = EXTRAPOLATION_MASKS

    # Apply shearing to the input image or mask
    if len(orig.shape) == 4:
        # Handle 3D images
        for n, slc in enumerate(orig):
            orig[n] = transform.warp(slc, inverse_map=afine_tf, order=final_order, preserve_range=True,
                                     mode='reflect')
        sheared_orig = orig
    else:
        sheared_orig = transform.warp(orig, inverse_map=afine_tf, order=final_order, preserve_range=True,
                                      mode='reflect')

    return sheared_orig


# allow for shift mask but only for the mask
# this cannot be an augmentation --> it is more complex than that
def mask_pixels_and_compute_for_those(orig, is_mask, shift_mask=False, **kwargs):
    """
    Mask pixels in the input image or mask and compute for those pixels.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        shift_mask (bool): A flag indicating whether to shift the mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The masked and computed image or mask.

    """
    if not is_mask:
        # Mask rows or columns in the input image
        rolled_orig = mask_rows_or_columns(orig, spacing_X=2, spacing_Y=5)
    else:
        if shift_mask:
            # Determine initial shifts for shifting the mask
            initial_shift_X = 0
            initial_shift_Y = 0
            if random.uniform(0, 1) < 0.5:
                if random.uniform(0, 1) < 0.5:
                    if random.uniform(0, 1) < 0.5:
                        initial_shift_X = 1
                    else:
                        initial_shift_Y = 1
                else:
                    initial_shift_X = 1
                    initial_shift_Y = 1

            # Mask rows or columns in the input mask with initial shifts
            rolled_orig = mask_rows_or_columns(orig, spacing_X=2, spacing_Y=5, initial_shiftX=initial_shift_X,
                                               initial_shiftY=initial_shift_Y, return_boolean_mask=True)
            orig[rolled_orig == False] = np.nan
            rolled_orig = orig
        else:
            # Mask rows or columns in the input mask without shifting
            rolled_orig = mask_rows_or_columns(orig, spacing_X=2, spacing_Y=5, return_boolean_mask=True)
            orig[rolled_orig == False] = np.nan
            rolled_orig = orig

    return rolled_orig


# rolls along the Z axis of a 3D image ignores for 2D images and ignores for masks
def rollZ(orig, is_mask, **kwargs):
    """
    Roll the input image or mask along the z-axis.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The rolled image or mask.

    # Examples:
    #     >>> image = np.array([[[1, 2, 3],
    #     ...                    [4, 5, 6],
    #     ...                    [7, 8, 9]],
    #     ...                   [[10, 11, 12],
    #     ...                    [13, 14, 15],
    #     ...                    [16, 17, 18]]])
    #     >>> result = rollZ(image, False)
    #     >>> print(result)
    #     [[[10 11 12]
    #       [13 14 15]
    #       [16 17 18]]
    #      [[1 2 3]
    #       [4 5 6]
    #       [7 8 9]]]
    """
    # Generate a random roll value for the z-axis
    random_roll = np.random.randint(0, orig.shape[0])
    if random_roll != 0:
        random_roll = random_roll if random.random() < 0.5 else -random_roll

    if len(orig.shape) != 4:
        # Do not perform roll for non-4D images
        random_roll = 0

    if random_roll != 0 and (not is_mask or (is_mask and len(orig.shape) == 4)) and len(orig.shape) == 4 and \
            random_roll != orig.shape[0] and random_roll != -orig.shape[0]:
        # Roll the image along the z-axis if conditions are met
        rolled_orig = np.roll(orig, random_roll, axis=0)
    else:
        # Keep the original image or mask
        rolled_orig = orig

    return rolled_orig


# shuffle images along the Z axis, may be useful for best focus algorithms
def shuffleZ(orig, is_mask, **kwargs):
    """
    Shuffle the input image or mask along the z-axis.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The shuffled image or mask.

    # Examples:
    #     >>> image = np.array([[[1, 2, 3],
    #     ...                    [4, 5, 6],
    #     ...                    [7, 8, 9]],
    #     ...                   [[10, 11, 12],
    #     ...                    [13, 14, 15],
    #     ...                    [16, 17, 18]]])
    #     >>> result = shuffleZ(image, False)
    #     >>> print(result)
    #     [[[10 11 12]
    #       [13 14 15]
    #       [16 17 18]]
    #      [[1 2 3]
    #       [4 5 6]
    #       [7 8 9]]]
    """
    if len(orig.shape) == 4 and (not is_mask or (is_mask and len(orig.shape) == 4)):
        # Shuffle the image along the z-axis
        np.random.shuffle(orig)

    return orig




# mega TODO --> allow flip 3D and if so check that masks that can be 2D remain ok
# in fact that should already work works
def flip(orig, is_mask, **kwargs):
    """
    Flip the input image or mask along a random axis.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The flipped image or mask.


    """
    if len(orig.shape) == 4:
        axis = random.choice([-2, -3, -4])  # Randomly choose an axis to flip
    else:
        axis = random.choice([-2, -3])

    if axis == -4 and len(orig.shape) == 3:
        # If the axis is -4 and the input is a 3D image, no flipping is performed
        flipped_orig = orig
    else:
        # Flip the image or mask along the chosen axis using np.flip
        flipped_orig = np.flip(orig, axis)

    return flipped_orig


def high_noise(orig, is_mask, **kwargs):
    """
    Add high levels of noise to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The image or mask with high levels of noise added.

    """
    return _add_noise(orig, is_mask, 'high')


def low_noise(orig, is_mask, **kwargs):
    """
    Add low levels of noise to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The image or mask with low levels of noise added.

    """
    return _add_noise(orig, is_mask, 'low')


# really a crappy augmenter probably not even worth keeping ????
def add_z_frames(orig, is_mask, z_frames_to_add_below=0, z_frames_to_add_above=0, **kwargs):
    """
    Add additional Z frames to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        z_frames_to_add_below (int): The number of Z frames to add below the original image or mask.
        z_frames_to_add_above (int): The number of Z frames to add above the original image or mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The image or mask with additional Z frames added.

    # Examples:
    #     >>> image = np.array([[[1, 2, 3],
    #     ...                    [4, 5, 6],
    #     ...                    [7, 8, 9]]])
    #     >>> result = add_z_frames(image, False, z_frames_to_add_below=1, z_frames_to_add_above=2)
    #     >>> print(result.shape)
    #     (3, 3, 5)
    """
    # Ignore for masks
    if is_mask and not len(orig.shape) == 4:
        return orig

    if z_frames_to_add_below == 0 and z_frames_to_add_above == 0:
        # Nothing to do
        return orig

    zpos = -4
    try:
        zpos = orig.get_dim_idx('d')
    except:
        logger.error('dimension d/z not found, assuming Z pos = -4, i.e. ...dhwc image')

    if orig.shape[zpos] == 1:
        # Image is 2D, nothing to do
        return orig

    if z_frames_to_add_above != 0:
        # Add Z frames before
        smallest_shape = list(orig.shape)
        smallest_shape[zpos] = z_frames_to_add_above
        missing_frames_above = np.zeros(smallest_shape, dtype=orig.dtype)

        orig = np.append(missing_frames_above, orig, axis=zpos)

    if z_frames_to_add_below != 0:
        # Add Z frames after
        smallest_shape[zpos] = z_frames_to_add_below
        missing_frames_below = np.zeros(smallest_shape, dtype=orig.dtype)

        orig = np.append(orig, missing_frames_below, axis=zpos)

    return orig


# use skimage to add noise
def _add_noise(orig, is_mask, strength='low'):
    """
    Add noise to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        strength (str): The strength of the noise. Valid values are 'low' or 'high'.

    Returns:
        ndarray: The image or mask with noise added.

    """
    mode = random.choice(['gaussian', 'localvar', 'poisson', 's&p', 'speckle'])

    if not is_mask:
        extra_params = {}
        if mode in ['salt', 'pepper', 's&p']:
            if strength == 'low':
                extra_params['amount'] = 0.1  # Weak amount of salt and pepper noise
            else:
                extra_params['amount'] = 0.25  # Strong amount of salt and pepper noise
        if mode in ['gaussian', 'speckle']:
            if strength == 'low':
                extra_params['var'] = 0.025  # Weak variance for Gaussian noise
            else:
                extra_params['var'] = 0.1  # Strong variance for Gaussian noise
            if mode == 'speckle':
                if strength == 'low':
                    extra_params['var'] = 0.1  # Weak variance for speckle noise
                else:
                    extra_params['var'] = 0.40  # Strong variance for speckle noise

        min = orig.min()
        max = orig.max()

        noisy_image = random_noise(orig, mode=mode, clip=True, **extra_params)

        if min == 0 and max == 1 or max == min:
            pass
        else:
            # We preserve the original image range
            noisy_image = (noisy_image * (max - min)) + min
    else:
        noisy_image = orig

    return noisy_image


def random_intensity_gamma_contrast_changer(orig, is_mask, **kwargs):
    """
    Apply random intensity, gamma, and contrast changes to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.

    Returns:
        ndarray: The image or mask with random intensity, gamma, and contrast changes applied.

    """
    return _random_intensity_gamma_contrast_changer(orig, is_mask)


# use skimage to change gamma/contrast, ... pixel intensity and/or scale
def _random_intensity_gamma_contrast_changer(orig, is_mask):
    """
    Apply random intensity, gamma, and contrast changes to the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.

    Returns:
        ndarray: The image or mask with random intensity, gamma, and contrast changes applied.

    """
    # fraction of image that should contain salt black or pepper white pixels
    mode = random.choice([exposure.rescale_intensity, exposure.adjust_gamma, exposure.adjust_log,
                          exposure.adjust_sigmoid])
    strength = random.choice(['medium', 'strong'])

    if not is_mask:
        if orig.min() >= 0:
            extra_params = {}
            if mode == exposure.rescale_intensity:
                if strength == 'strong':
                    v_min, v_max = np.percentile(orig, (5, 95))
                else:
                    v_min, v_max = np.percentile(orig, (0.9, 98))
                extra_params['in_range'] = (v_min, v_max)
            elif mode == exposure.adjust_gamma:
                if strength == 'strong':
                    extra_params['gamma'] = 0.4
                    extra_params['gain'] = 0.9
                else:
                    extra_params['gamma'] = 0.8
                    extra_params['gain'] = 0.9
            elif mode == exposure.adjust_sigmoid:
                if strength == 'strong':
                    extra_params['gain'] = 5
                else:
                    extra_params['gain'] = 2

            contrast_changed_image = mode(orig, **extra_params)
        else:
            logger.warning(
                'Negative intensities detected --> ignoring random_intensity_gamma_contrast_changer augmentation.')
            contrast_changed_image = orig
    else:
        contrast_changed_image = orig

    return contrast_changed_image


def change_image_intensity_and_shift_range(orig, is_mask, **kwargs):
    """
    Change the intensity and shift the range of the input image or mask.

    Args:
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The image or mask with changed intensity and shifted range.

    """
    scaling_factor = random.uniform(kwargs['intensity'], 1)

    for c in range(orig.shape[-1]):
        cur = orig[..., c]
        if not is_mask:
            min_before = cur.min()
            initial_range = cur.max() - min_before

            try:
                import numexpr
                cur = numexpr.evaluate("cur * scaling_factor")
            except:
                cur = cur * scaling_factor

            new_min = cur.min()
            shift_min = min_before - new_min

            try:
                import numexpr
                cur = numexpr.evaluate("cur + shift_min")
            except:
                cur += shift_min

        shift_range = bool(random.getrandbits(1))
        if shift_range:
            if not is_mask:
                new_range = cur.max() - cur.min()
                possible_range_increase = initial_range - new_range
                random_range_shift = random.uniform(0., possible_range_increase)

                try:
                    import numexpr
                    cur = numexpr.evaluate("cur + random_range_shift")
                except:
                    cur += random_range_shift

        orig[..., c] = cur

    out = orig

    if is_mask:
        out = orig

    return out


def execute_chained_augmentations(chained_augmentations, orig, is_mask, **kwargs):
    """
    Execute a chain of augmentations on the input image or mask.

    Args:
        chained_augmentations (list): A list of augmentation functions to apply.
        orig (ndarray): The input image or mask.
        is_mask (bool): A flag indicating whether the input is a mask.
        **kwargs: Additional keyword arguments.

    Returns:
        ndarray: The augmented image or mask.

    """

    from epyseg.deeplearning.augmentation.generators.data import DataGenerator

    if chained_augmentations is None:
        logger.error('No augmentation in the chain. Ignoring.')
        return orig

    if not isinstance(chained_augmentations, list):
        chained_augmentations = [chained_augmentations]

    for aug in chained_augmentations:
        if aug is None:
            continue
        if isinstance(aug, str):
            if aug.lower() == 'none':
                continue
            if aug in DataGenerator.augmentation_types_and_methods:
                aug = DataGenerator.augmentation_types_and_methods[aug]

        orig = aug(orig, is_mask=is_mask, **kwargs)

    return orig



def tst_all_augs():
    """
    A basic test method for all augs.

    # Examples:
    #     >>> tst_all_augs()
    """
    from epyseg.deeplearning.augmentation.generators.data import DataGenerator
    # TODO --> could do all the tests here and check that they are always in sync despite not being used ---> can I call random after to check in sync --> yes -> very smart anc non costly idea
    orig = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012.png')[..., 1]
    mask = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/handCorrection.tif')

    # try with wings
    orig = Img('/E/Sample_images/adult_wings/pigmented_wings_benjamin_all/scaled_half/N01_YwrGal4_males_wing0_projected.tif')
    mask = Img('/E/Sample_images/adult_wings/pigmented_wings_benjamin_all/scaled_half/predict/N01_YwrGal4_males_wing0_projected.tif')[...,0]
    # print(orig.shape, mask.shape)

    orig_3D = Img('/E/Sample_images/Amrutha_gut/trash_output_tests/test_2D_to_3D/orig.tif')
    mask_3D = Img('/E/Sample_images/Amrutha_gut/trash_output_tests/test_2D_to_3D/segmented_gut.tif')

    # we fake a set of parameters here

    augmentation_types_and_ranges = DataGenerator.augmentation_types_and_ranges
    augmentation_types_and_values =  DataGenerator.augmentation_types_and_values

    if True:
        from personal.ensemble.preview_many_images_side_by_side import preview_as_panel
        # raise Exception('THERE IS A BUG HERE IN THE RANGE OF THE MASK --> HOW CAN I CHANGE THIS --> TODO')
        print('testing crop')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)

        crop_with_parameters = partial(crop_augmentation, coords_dict={'x1': 10, 'x2': -10, 'y1': 10,
                                                                       'y2': -10})  # does that work with negative values by the way

        # NB image need hwc
        orig_2D_plus = orig[..., np.newaxis]
        augmented_orig = crop_with_parameters(orig=np.copy(orig_2D_plus), is_mask=False,
                                              **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        # NB image need hwc
        mask_2D_plus = mask[..., np.newaxis]
        augmented_mask = crop_with_parameters(orig=np.copy(mask_2D_plus), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        # that does not work --> but why is that
        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask_2D_plus, augmented_mask, success_message='mask is augmented')
        # import sys
        # sys.exit(0)

    if True:

        # raise Exception('THERE IS A BUG HERE IN THE RANGE OF THE MASK --> HOW CAN I CHANGE THIS --> TODO')
        print('testing elastic_with_zoom_and_translation')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = elastic_with_zoom_and_translation(orig=np.copy(orig), is_mask=False,
                                                           **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = elastic_with_zoom_and_translation(orig=np.copy(mask), is_mask=True,
                                                           **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        # that does not work --> but why is that
        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask, augmented_mask, success_message='mask is augmented')
        # import sys
        # sys.exit(0)

    if True:
        # raise Exception('THERE IS A BUG HERE IN THE RANGE OF THE MASK --> HOW CAN I CHANGE THIS --> TODO')
        print('testing elastic')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = elastic(orig=np.copy(orig), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = elastic(orig=np.copy(mask), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        # that does not work --> but why is that
        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask, augmented_mask, success_message='mask is augmented')
        # import sys
        # sys.exit(0)

    if True:
        print('testing change_image_intensity_and_shift_range')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        # NB image need hwc
        orig_2D_plus = orig[..., np.newaxis]
        augmented_orig = change_image_intensity_and_shift_range(orig=np.copy(orig_2D_plus), is_mask=False,
                                                                **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        # NB image need hwc
        mask_2D_plus = mask[..., np.newaxis]
        augmented_mask = change_image_intensity_and_shift_range(orig=np.copy(mask_2D_plus), is_mask=True,
                                                                **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        # that does not work --> but why is that
        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_equal(mask_2D_plus, augmented_mask, success_message='mask was kept unchanged')
        # import sys
        # sys.exit(0)

    if True:
        print('testing augmentation of original')
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = blur(orig=np.copy(orig), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = blur(orig=np.copy(mask), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_equal(mask, augmented_mask, success_message='mask is kept unchanged')

    if True:
        print('testing graded_intensity_modification')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = graded_intensity_modification(orig=np.copy(orig), is_mask=False,
                                                       **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = graded_intensity_modification(orig=np.copy(mask), is_mask=True,
                                                       **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_equal(mask, augmented_mask, success_message='mask is kept unchanged')

    if True:
        print('testing invert')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = invert(orig=np.copy(orig), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = invert(orig=np.copy(mask), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_equal(mask, augmented_mask, success_message='mask is kept unchanged')

    if True:
        print('testing rotate')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        # NB image need dhwc
        orig_2D_plus = orig[..., np.newaxis]
        augmented_orig = rotate(orig=np.copy(orig_2D_plus), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        # NB image need dhwc
        mask_2D_plus = mask[..., np.newaxis]
        augmented_mask = rotate(orig=np.copy(mask_2D_plus), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is changed
        array_different(mask_2D_plus, augmented_mask, success_message='mask was augmented')

    if True:
        print('testing rotate_interpolation_free')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        # NB image need hwc
        orig_2D_plus = orig[..., np.newaxis]
        augmented_orig = rotate_interpolation_free(orig=np.copy(orig_2D_plus), is_mask=False,
                                                   **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        #  NB image need hwc
        mask_2D_plus = mask[..., np.newaxis]
        augmented_mask = rotate_interpolation_free(orig=np.copy(mask_2D_plus), is_mask=True,
                                                   **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask_2D_plus, augmented_mask, success_message='mask has been augmented')

    if True:
        print('testing translate')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        #  NB image need hwc
        orig_2D_plus = orig[..., np.newaxis]
        augmented_orig = translate(orig=np.copy(orig_2D_plus), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        #  NB image need hwc
        mask_2D_plus = mask[..., np.newaxis]
        augmented_mask = translate(orig=np.copy(mask_2D_plus), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask_2D_plus, augmented_mask, success_message='mask was augmented')

    if True:
        print('testing zoom')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = zoom(orig=np.copy(orig), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = zoom(orig=np.copy(mask), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask, augmented_mask, success_message='mask was augmented')

    if True:
        print('testing shear')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = shear(orig=np.copy(orig), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = shear(orig=np.copy(mask), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask, augmented_mask, success_message='mask was augmented')

    if True:
        # NB THIS IS NOT A DATA AUG !!!!!!
        print('testing mask_pixels_and_compute_for_those')
        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        # assumes image si à minima hwc
        #  NB image need hwc
        orig_2D_plus = orig[..., np.newaxis]
        augmented_orig = mask_pixels_and_compute_for_those(orig=np.copy(orig_2D_plus), is_mask=False,
                                                           **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        # NB image needs be float for it to work and this is definitely not a data augmentation --> do not put it in data aug but rather in post processing and can apply it to compute losses
        mask_2D_plus = mask[..., np.newaxis]
        augmented_mask = mask_pixels_and_compute_for_those(orig=np.copy(mask_2D_plus).astype(float), is_mask=True,
                                                           **augmentation_types_and_values)
        augmented_mask = augmented_mask.astype(bool)  # nb it does return a True or False mask
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask_2D_plus, augmented_mask, success_message='mask was augmented')

    if True:
        print('testing rollZ')
        # testing augmentation of original
        # NB image need dhwc
        orig_3D_plus = orig_3D[..., np.newaxis]
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = rollZ(orig=np.copy(orig_3D_plus), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        # NB image need dhwc
        mask_3D_plus = mask_3D[..., np.newaxis]
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = rollZ(orig=np.copy(mask_3D_plus), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        mask_2D = mask[..., np.newaxis]
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask2 = rollZ(orig=np.copy(mask_2D), is_mask=True, **augmentation_types_and_values)
        val_after_third = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second == val_after_third, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            try:
                # TODO --> maybe plot both side by side
                plt.imshow(augmented_mask2)
                plt.show()
                # plt.imshow(augmented_mask)
                # plt.show()
                # plt.imshow(augmented_orig)
                # plt.show()
                preview_as_panel([augmented_orig, augmented_mask])
            except:
                # plt.imshow(augmented_mask[int((augmented_mask.shape[0])/2)])
                # plt.show()
                # plt.imshow(augmented_orig[int((augmented_orig.shape[0])/2)])
                # plt.show()
                preview_as_panel([augmented_orig[int((augmented_orig.shape[0]) / 2)],
                                  augmented_mask[int((augmented_mask.shape[0]) / 2)]])
                # preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)
        print('augmented_mask2', augmented_mask2.max(), augmented_mask2.min(), augmented_mask2.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_3D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask_3D_plus, augmented_mask, success_message='mask  was augmented')

        array_equal(mask_2D, augmented_mask2, success_message='2D mask was unchanged')

    if True:
        print('testing shuffleZ')
        # testing augmentation of original
        # NB image need dhwc
        orig_3D_plus = orig_3D[..., np.newaxis]
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = shuffleZ(orig=np.copy(orig_3D_plus), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        # NB image need dhwc
        mask_3D_plus = mask_3D[..., np.newaxis]
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = shuffleZ(orig=np.copy(mask_3D_plus), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        mask_2D = mask[..., np.newaxis]
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask2 = shuffleZ(orig=np.copy(mask_2D), is_mask=True, **augmentation_types_and_values)
        val_after_third = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second == val_after_third, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            try:
                # TODO --> maybe plot both side by side
                plt.imshow(augmented_mask2)
                plt.show()
                # plt.imshow(augmented_mask)
                # plt.show()
                # plt.imshow(augmented_orig)
                # plt.show()
                preview_as_panel([augmented_orig, augmented_mask])
            except:
                # plt.imshow(augmented_mask[int((augmented_mask.shape[0])/2)])
                # plt.show()
                # plt.imshow(augmented_orig[int((augmented_orig.shape[0])/2)])
                # plt.show()
                preview_as_panel([augmented_orig[int((augmented_orig.shape[0]) / 2)],
                                  augmented_mask[int((augmented_mask.shape[0]) / 2)]])
                # preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)
        print('augmented_mask2', augmented_mask2.max(), augmented_mask2.min(), augmented_mask2.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_3D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask_3D_plus, augmented_mask, success_message='mask  was augmented')

        array_equal(mask_2D, augmented_mask2, success_message='2D mask was unchanged')

    if True:
        print('testing flip')
        # testing augmentation of original
        # NB image need dhwc
        orig_2D_plus = orig[..., np.newaxis]
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = flip(orig=np.copy(orig_2D_plus), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        # NB image need dhwc
        mask_2D_plus = mask[..., np.newaxis]
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = flip(orig=np.copy(mask_2D_plus), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask_2D_plus, augmented_mask, success_message='mask was augmented')

    if True:
        # NB why is max noise 254 and not 255 --> MEGA TODO CHECK --> did i do a mistake ????
        print('testing high_noise')
        # testing augmentation of original
        # NB image need dhwc
        orig_2D_plus = orig[..., np.newaxis]
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = high_noise(orig=np.copy(orig_2D_plus), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        # NB image need dhwc
        mask_2D_plus = mask[..., np.newaxis]
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = high_noise(orig=np.copy(mask_2D_plus), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_equal(mask_2D_plus, augmented_mask, success_message='mask was not augmented')

    if True:
        print('testing low_noise')
        # testing augmentation of original
        # NB image need dhwc
        orig_2D_plus = orig[..., np.newaxis]
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = low_noise(orig=np.copy(orig_2D_plus), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        # NB image need dhwc
        mask_2D_plus = mask[..., np.newaxis]
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = low_noise(orig=np.copy(mask_2D_plus), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_equal(mask_2D_plus, augmented_mask, success_message='mask was not augmented')

    if True:
        print('testing random_intensity_gamma_contrast_changer')
        # testing augmentation of original
        # NB image need dhwc
        orig_2D_plus = orig[..., np.newaxis]
        seed = datetime.now()
        random.seed(seed)
        augmented_orig = random_intensity_gamma_contrast_changer(orig=np.copy(orig_2D_plus), is_mask=False,
                                                                 **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        # NB image need dhwc
        mask_2D_plus = mask[..., np.newaxis]
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = random_intensity_gamma_contrast_changer(orig=np.copy(mask_2D_plus), is_mask=True,
                                                                 **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_equal(mask_2D_plus, augmented_mask, success_message='mask was not augmented')

    # seems ok but could do the test more smartly by checking that the size of some of the dimensions matches the expected
    # this aug anyways sucks!!!
    # maybe add a parameter force mask in some cases because that could be useful --> but what I do is probably ok still in fact !!!
    if True:
        print('testing add_z_frames')
        # testing augmentation of original
        # NB image need dhwc
        orig_3D_plus = orig_3D[..., np.newaxis]
        seed = datetime.now()
        random.seed(seed)

        # the values should be passed through the stuff
        augmentation_types_and_values['z_frames_to_add_below'] = 2
        augmentation_types_and_values['z_frames_to_add_above'] = 4

        augmented_orig = add_z_frames(orig=np.copy(orig_3D_plus), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        # NB image need dhwc
        mask_3D_plus = mask_3D[..., np.newaxis]
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask = add_z_frames(orig=np.copy(mask_3D_plus), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        mask_2D = mask[..., np.newaxis]
        random.seed(seed)  # reseed to be sure the same random are applied
        augmented_mask2 = shuffleZ(orig=np.copy(mask_2D), is_mask=True, **augmentation_types_and_values)
        val_after_third = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second == val_after_third, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if False:
            try:
                # TODO --> maybe plot both side by side
                plt.imshow(augmented_mask2)
                plt.show()
                # plt.imshow(augmented_mask)
                # plt.show()
                # plt.imshow(augmented_orig)
                # plt.show()
                preview_as_panel([augmented_orig, augmented_mask])
            except:
                # plt.imshow(augmented_mask[int((augmented_mask.shape[0])/2)])
                # plt.show()
                # plt.imshow(augmented_orig[int((augmented_orig.shape[0])/2)])
                # plt.show()
                preview_as_panel([augmented_orig[int((augmented_orig.shape[0]) / 2)],
                                  augmented_mask[int((augmented_mask.shape[0]) / 2)]])
                # preview_as_panel([augmented_orig, augmented_mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)
        print('augmented_mask2', augmented_mask2.max(), augmented_mask2.min(), augmented_mask2.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_3D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        array_different(mask_3D_plus, augmented_mask, success_message='mask was augmented')

        array_equal(mask_2D, augmented_mask2, success_message='2D mask was unchanged')


if __name__ == '__main__':

    from epyseg.deeplearning.augmentation.generators.data import DataGenerator
    # TODO --> could do all the tests here and check that they are always in sync despite not being used ---> can I call random after to check in sync --> yes -> very smart anc non costly idea
    orig = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012.png')[..., 1]
    mask = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/handCorrection.tif')

    # try with wings
    orig = Img('/E/Sample_images/adult_wings/pigmented_wings_benjamin_all/scaled_half/N01_YwrGal4_males_wing0_projected.tif')
    mask = Img('/E/Sample_images/adult_wings/pigmented_wings_benjamin_all/scaled_half/predict/N01_YwrGal4_males_wing0_projected.tif')[...,0]
    # print(orig.shape, mask.shape)

    orig_3D = Img('/E/Sample_images/Amrutha_gut/trash_output_tests/test_2D_to_3D/orig.tif')
    mask_3D = Img('/E/Sample_images/Amrutha_gut/trash_output_tests/test_2D_to_3D/segmented_gut.tif')

    # we fake a set of parameters here

    # TODO should I also store increment there ???
    augmentation_types_and_ranges = DataGenerator.augmentation_types_and_ranges
    augmentation_types_and_values = DataGenerator.augmentation_types_and_values

    # if True:
    #     preview_as_panel([orig, mask])
    #     import sys
    #     sys.exit(0)


    # if True:
    #     # they are not the same so I need to seed them
    #     import sys
    #     seed = datetime.now()
    #     random.seed(seed)
    #     np.random.seed(random.randint(0, 10000000))
    #     displacement1 = np.random.randn(2, 3, 3) * 25
    #     random.seed(seed)
    #     np.random.seed(random.randint(0, 10000000))
    #     displacement2 = np.random.randn(2, 3, 3) * 25
    #     array_equal(displacement1, displacement2,success_message='arrays are the same')
    #     import sys
    #     sys.exit(0)

    if False:
        # in fact elastic deform is super weak for big images such as wings --> how can I increase its force so that it becomes a useful augmentation

        # orig = np.random.rand(1289,2001)
        orig = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012.png')[..., 1] #strong deformation
        orig = Img('/E/Sample_images/adult_wings/pigmented_wings_benjamin_all/scaled_half/N01_YwrGal4_males_wing0_projected.tif') #very weak deformation
        augmented_orig = strong_elastic_with_zoom_and_translation(orig, False, **augmentation_types_and_values)
        preview_as_panel([augmented_orig,  orig])

        import sys
        sys.exit(0)


    if True:
        from personal.ensemble.preview_many_images_side_by_side import preview_as_panel
        # that works
        # try pipes with string names --> can I do that ??? --> probably yes
        # it seems to work --> now just plot them side by side --> try that --> TODO
        print('testing piped data aug')

        # parfait tt marche --> cool --> implementer ça un peu partout maintenant et voir ce que ça fait si je le passe à un data aug --> à faire

        # chained_augmentation = [pad_border_xy, invert, rotate_interpolation_free, flip, elastic]  #, elastic] --> elastic does not work --> it fucks the pipe --> how can I do that then ????
        crop_with_parameters = partial(crop_augmentation, coords_dict={'x1':10, 'x2':-10, 'y1':10, 'y2':-10}) # does that work with negative values by the way
        chained_augmentation = [crop_with_parameters, strong_elastic]  # elastic_with_zoom_and_translation  #, elastic] --> elastic does not work --> it fucks the pipe --> how can I do that then ????
        # chained_augmentation = [pad_border_xy, 'invert', 'rotate (interpolation free)', 'flip', 'elastic']  #, elastic] --> elastic does not work --> it fucks the pipe --> how can I do that then ????

        # testing augmentation of original
        seed = datetime.now()
        random.seed(seed)
        # NB image need dhwc
        orig_2D_plus = orig[..., np.newaxis]
        # pb is how to pass the immutable params to all the functions --> probably need make my own function here once for good
        augmented_orig = execute_chained_augmentations(chained_augmentation, orig=np.copy(orig_2D_plus), is_mask=False, **augmentation_types_and_values) #blur(orig=np.copy(orig), is_mask=False, **augmentation_types_and_values)
        val_after_first = random.random()  # will be used to check that random is still in sync after the call

        # testing augmentation/non-augmentation of mask
        random.seed(seed)  # reseed to be sure the same random are applied
        # NB image need hwc
        mask_2D_plus = mask[..., np.newaxis]
        augmented_mask = execute_chained_augmentations(chained_augmentation,  orig=np.copy(mask_2D_plus), is_mask=True, **augmentation_types_and_values)
        val_after_second = random.random()  # will be used to check that random is still in sync after the call

        # try:
        assert val_after_first == val_after_second, "NB Random is NOT IN SYNC, the nb of random calls is not equal"
        print('EVERYTHING OK: Random is in SYNC')
        #        except AssertionError as e:
        #            print(e)

        if True:
            # TODO --> maybe plot both side by side
            # plt.imshow(augmented_mask)
            # plt.show()
            # plt.imshow(augmented_orig)
            # plt.show()
            preview_as_panel([augmented_orig, augmented_mask, orig, mask])

        print('orig', orig.max(), orig.min(), orig.shape)
        print('mask', mask.max(), mask.min(), mask.shape)

        print('augmented_orig', augmented_orig.max(), augmented_orig.min(), augmented_orig.shape)
        print('augmented_mask', augmented_mask.max(), augmented_mask.min(), augmented_mask.shape)

        # asserts changes were applied to the original (is that smart ???) --> since it's random and value can be 0 then may make no sense but most often it will be a valid check
        array_different(orig_2D_plus, augmented_orig, success_message='original image was augmented')

        # assert the mask is unchanged
        try:
            array_equal(mask_2D_plus, augmented_mask, success_message='mask is kept unchanged')
        except:
            array_different(mask_2D_plus, augmented_mask, success_message='mask is changed')

        import sys
        sys.exit(0)


import numpy as np
from functools import partial
# from PIL import Image
from scipy.ndimage import generic_filter
from epyseg.img import Img, pad_border_xy
import matplotlib.pyplot as plt

def associate_vertices_to_cells(cell_labels, vertices, forbidden_colors=(0, 0xFFFFFF),
                                output_vertices_and_associated_cells=False, output_cells_and_their_vertices=False,
                                output_cells_and_their_neighbors=False):
    """
    Associates vertices with corresponding cells based on cell labels.

    Args:
        cell_labels (numpy.ndarray): Array containing cell labels.
        vertices (tuple or numpy.ndarray): Tuple of vertex coordinates or vertex array.
        forbidden_colors (tuple, optional): Colors to ignore when associating vertices to cells. Defaults to (0, 0xFFFFFF).
        output_vertices_and_associated_cells (bool, optional): Whether to output vertices and their associated cells. Defaults to False.
        output_cells_and_their_vertices (bool, optional): Whether to output cells and their associated vertices. Defaults to False.
        output_cells_and_their_neighbors (bool, optional): Whether to output cells and their neighboring cells. Defaults to False.

    Returns:
        dict or list: Output based on the selected options.

    """
    outputs = []
    cell_id_n_vx_coords = {}
    vertices_n_associated_cells = {}

    if isinstance(vertices, tuple):
        vx_coords = vertices
    else:
        vx_coords = np.where(vertices == 255)

    for i in range(len(vx_coords[0])):
        y = vx_coords[0][i]
        x = vx_coords[1][i]
        ids = neighbors8((y, x), cell_labels)

        ids = set(ids)

        if forbidden_colors:
            for col in forbidden_colors:
                if col in ids:
                    ids.remove(col)

        vertices_n_associated_cells[(y, x)] = ids

        for id in ids:
            if id in cell_id_n_vx_coords:
                cell_id_n_vx_coords[id].append((y, x))
            else:
                cell_id_n_vx_coords[id] = [(y, x)]

    if output_cells_and_their_vertices:
        outputs.append(cell_id_n_vx_coords)

    if output_vertices_and_associated_cells:
        outputs.append(vertices_n_associated_cells)

    if output_cells_and_their_neighbors:
        cell_ids_n_neighbor_ids = {}
        for key, value in cell_id_n_vx_coords.items():
            neighbs = []
            for vx in value:
                neighbs += vertices_n_associated_cells[vx]

            neighbs = set(neighbs)
            neighbs.remove(key)

            cell_ids_n_neighbor_ids[key] = neighbs

        outputs.append(cell_ids_n_neighbor_ids)

    if len(outputs) == 1:
        return outputs[0]
    else:
        if outputs:
            return outputs

    return cell_id_n_vx_coords


def detect_neighbors_of_current_cell(cell_id, cell_vertices, cell_labels, forbidden_colors=(0,0xFFFFFF)):
    """
    Detects the neighbors of a current cell based on its vertices.

    Args:
        cell_id (int): ID of the cell.
        cell_vertices (list): List of cell vertices.
        cell_labels (numpy.ndarray): Array containing cell labels.
        forbidden_colors (tuple, optional): Colors to ignore when detecting neighbors. Defaults to (0, 0xFFFFFF).

    Returns:
        set: Set of neighboring cell IDs.

    """
    neighbs = []
    for vx in cell_vertices:
        neighbs += neighbors8(vx, cell_labels).tolist()

    neighbs = set(neighbs)
    neighbs.remove(cell_id)

    if forbidden_colors:
        for col in forbidden_colors:
            if col in neighbs:
                neighbs.remove(col)

    return neighbs

def neighbors8(vx_coords, cell_labels):
    """
    Calculates the 8 neighbors surrounding a given vertex in a matrix of cell labels.

    Args:
        vx_coords (tuple): The coordinates of the vertex (y, x).
        cell_labels (ndarray): The matrix of cell labels.

    Returns:
        ndarray: The cell labels of the neighboring cells as a 1D array.

    """
    min_y = vx_coords[0] - 1
    max_y = vx_coords[0] + 2
    min_x = vx_coords[1] - 1
    max_x = vx_coords[1] + 2

    min_y = min_y if min_y >= 0 else 0
    max_y = max_y if max_y < cell_labels.shape[0] else cell_labels.shape[0]
    min_x = min_x if min_x >= 0 else 0
    max_x = max_x if max_x < cell_labels.shape[1] else cell_labels.shape[1]

    return cell_labels[min_y:max_y, min_x:max_x].ravel()


def neighbors8_shift_mapping(vx_coords, cell_labels):
    """
    Calculates the coordinates of the 8 neighbors surrounding a given vertex in a matrix of cell labels.

    Args:
        vx_coords (tuple): The coordinates of the vertex (y, x).
        cell_labels (ndarray): The matrix of cell labels.

    Returns:
        list: The coordinates of the neighboring vertices.

    """
    min_y = vx_coords[0] - 1
    max_y = vx_coords[0] + 2
    min_x = vx_coords[1] - 1
    max_x = vx_coords[1] + 2

    min_y = min_y if min_y >= 0 else 0
    max_y = max_y if max_y < cell_labels.shape[0] else cell_labels.shape[0]
    min_x = min_x if min_x >= 0 else 0
    max_x = max_x if max_x < cell_labels.shape[1] else cell_labels.shape[1]

    coords = []
    for y in range(min_y, max_y):
        for x in range(min_x, max_x):
            coords.append([y, x])

    return coords

def cantor_color_bonds(RGB24_or_lab, vertices, boundary_color=0xFFFFFF, copy=True):
    """
    Applies the Cantor pairing function to color the bonds between vertices in an image.

    Args:
        RGB24_or_lab (ndarray): The input image in RGB or LAB format.
        vertices (ndarray): The matrix of vertices in the image.
        boundary_color (int or list): The color(s) used for boundary detection. Defaults to 0xFFFFFF.
        copy (bool): Specifies whether to create a copy of the image. Defaults to True.

    Returns:
        ndarray: The matrix of colored bonds.

    """
    if isinstance(boundary_color, int):
        boundary_color = [boundary_color]

    method = partial(_cantor_color_bonds, boundary_colors=boundary_color)

    if copy:
        RGB24 = np.copy(RGB24_or_lab)
    else:
        RGB24 = RGB24_or_lab

    bonds = generic_filter(RGB24, method, (3, 3))
    bonds[vertices != 0] = 0

    max_id = bonds.max() + 1

    for iii in range(RGB24.shape[1]):
        if vertices[0, iii] == 255:
            max_id += 1
            continue
        bonds[0, iii] = max_id

    for iii in range(RGB24.shape[1]):
        if vertices[RGB24.shape[0] - 1, iii] == 255:
            max_id += 1
            continue
        bonds[RGB24.shape[0] - 1, iii] = max_id

    for jjj in range(RGB24.shape[0]):
        if vertices[jjj, 0] == 255:
            max_id += 1
            continue
        bonds[jjj, 0] = max_id

    for jjj in range(RGB24.shape[0]):
        if vertices[jjj, RGB24.shape[1] - 1] == 255:
            max_id += 1
            continue
        bonds[jjj, RGB24.shape[1] - 1] = max_id

    return bonds


def detect_vertices_and_bonds(RGB24_or_lab, detect_bonds=False,
                              boundary_color=0xFFFFFF, split_bonds_and_vertices=False, copy=True,
                              pad_to_detect_vertices=True):
    """
    Detects vertices and bonds in an image.

    Args:
        RGB24_or_lab (ndarray): The input image in RGB or LAB format.
        detect_bonds (bool): Specifies whether to detect bonds. Defaults to False.
        boundary_color (int or list): The color(s) used for boundary detection. Defaults to 0xFFFFFF.
        split_bonds_and_vertices (bool): Specifies whether to split bonds and vertices into separate matrices.
            Defaults to False.
        copy (bool): Specifies whether to create a copy of the image. Defaults to True.
        pad_to_detect_vertices (bool): Specifies whether to pad the image for vertex detection. Defaults to True.

    Returns:
        ndarray or tuple: The matrix of vertices, or a tuple of matrices if split_bonds_and_vertices is True.

    """
    if isinstance(boundary_color, int):
        boundary_color = [boundary_color]

    method = partial(_detect_vertices_and_bonds, boundary_colors=boundary_color)

    if copy and not pad_to_detect_vertices:
        RGB24 = np.copy(RGB24_or_lab)
    else:
        RGB24 = RGB24_or_lab

    if pad_to_detect_vertices:
        RGB24 = np.pad(RGB24, pad_width=1, mode='constant', constant_values=False)
        RGB24 = pad_border_xy(RGB24, mode=RGB24.max() + 1)

    vertices = generic_filter(RGB24, method, (3, 3))

    if pad_to_detect_vertices:
        vertices = vertices[1:-1, 1:-1]
        vertices[0, 0] = 255
        vertices[-1, 0] = 255
        vertices[0, -1] = 255
        vertices[-1, -1] = 255

        RGB24 = RGB24[1:-1, 1:-1]
    else:
        for jjj in range(1, RGB24.shape[0] - 1):
            neighbs = RGB24[jjj - 1:jjj + 2, 0:2]
            different_neighbours = set(neighbs.ravel().tolist())
            if len(different_neighbours) == 3:
                vertices[jjj][0] = 255
            neighbs = RGB24[jjj - 1:jjj + 2, RGB24.shape[0] - 2:RGB24.shape[0]]
            different_neighbours = set(neighbs.ravel().tolist())
            if len(different_neighbours) == 3:
                vertices[jjj][RGB24.shape[0] - 1] = 255

        for iii in range(1, RGB24.shape[1] - 1):
            neighbs = RGB24[0:2, iii - 1:iii + 2]
            different_neighbours = set(neighbs.ravel().tolist())
            if len(different_neighbours) == 3:
                vertices[0][iii] = 255
            neighbs = RGB24[RGB24.shape[1] - 2:RGB24.shape[1], iii - 1:iii + 2]
            different_neighbours = set(neighbs.ravel().tolist())
            if len(different_neighbours) == 3:
                vertices[RGB24.shape[0] - 1][iii] = 255

        vertices[0][0] = 255
        vertices[RGB24.shape[0] - 1][0] = 255
        vertices[RGB24.shape[0] - 1][RGB24.shape[1] - 1] = 255
        vertices[0][RGB24.shape[1] - 1] = 255

    if detect_bonds:
        vertices[np.where((RGB24 == boundary_color) & (vertices != 255))] = 128

    if split_bonds_and_vertices and detect_bonds:
        bonds = np.zeros_like(vertices)
        bonds[vertices == 128] = 128
        vertices[vertices == 128] = 0
        return vertices, bonds

    return vertices


def _cantor_pairing(id1, id2, _sort_points=True):
    """
    Applies the Cantor pairing function to two IDs to generate a unique bond ID.

    Args:
        id1 (int): The first ID.
        id2 (int): The second ID.
        _sort_points (bool): Specifies whether to sort the IDs before applying the function. Defaults to True.

    Returns:
        int: The bond ID.

    Examples:
        >>> _cantor_pairing(3, 4)
        31
        >>> _cantor_pairing(10, 5)
        125
    """
    x = id1
    y = id2
    if _sort_points and id1 < id2:
        x = id2
        y = id1
    return int(((x + y) * (x + y + 1) / 2) + y)

def _detect_vertices_and_bonds_old_doing_useless_stuff(P, detect_bonds=False, boundary_colors=(0xFFFFFF, 0)):
    if boundary_colors:
        if P[4] not in boundary_colors:  # not pure white
            return 0

    size = len(set(P))  # 4-way vertices only but why not
    if detect_bonds and size == 3:
        return 128
    return 255 if size >= 4 else 0


def _detect_vertices_and_bonds(P, boundary_colors=(0xFFFFFF, 0)):
    """
    Detects vertices in a neighborhood and assigns IDs to them.

    Args:
        P (ndarray): Neighborhood array.
        boundary_colors (tuple): Tuple of boundary colors. Defaults to (0xFFFFFF, 0).

    Returns:
        int: The vertex ID.

    """
    if boundary_colors:
        if P[4] not in boundary_colors:  # not pure white
            return 0

    size = len(set(P))  # 4-way vertices only but why not
    return 255 if size >= 4 else 0


def _identify_vertices_and_bonds(P, detect_bonds=False, boundary_colors=(0xFFFFFF, 0)):
    """
    Identifies vertices and assigns IDs to them. Optionally detects bonds.

    Args:
        P (ndarray): Neighborhood array.
        detect_bonds (bool): Specifies whether to detect bonds. Defaults to False.
        boundary_colors (tuple): Tuple of boundary colors. Defaults to (0xFFFFFF, 0).

    Returns:
        int: The vertex ID or bond ID.

    """
    if boundary_colors:
        if P[4] not in boundary_colors:  # not pure white
            return 0

    corrected_ids = set(P)
    size = len(corrected_ids)  # 4-way vertices only but why not

    if detect_bonds and size == 3:
        corrected_ids = list(corrected_ids.difference(boundary_colors))
        return _cantor_pairing(corrected_ids[0], corrected_ids[1])

    return 255 if size >= 4 else 0


def _cantor_color_bonds(P, boundary_colors=(0xFFFFFF, 0)):
    if boundary_colors:
        if P[4] not in boundary_colors:  # not pure white
            return 0

    corrected_ids = set(P)
    size = len(corrected_ids)  # 4-way vertices only but why not

    if size == 3:
        corrected_ids = list(corrected_ids.difference(boundary_colors))
        return _cantor_pairing(corrected_ids[0], corrected_ids[1])

    return 0


if __name__ == '__main__':

    if False:
        import sys
        print(_cantor_pairing(0xFFFFFF,0xFFFFFE)) #562949382996104 562949886312449
        print(_cantor_pairing(0xFFFFFF,0xFFFFFF)) #562949382996104 562949886312449
        print(_cantor_pairing(0xFFFFFFFF,0xFFFFFFFE)) #3.6893488130239234e+19 --> very big
        print(_cantor_pairing(0xFFFFFFFF,0xFFFFFFFF)) #3.689348813882917e+19 --> very big too # try avoid such big number but ok if can't do better...
        print(_cantor_pairing(0, 0))
        print(_cantor_pairing(0, 1)) # 0,1 and 1, 0 differ --> need sort the points before passing them or after passing them
        print(_cantor_pairing(1, 0))
        print(_cantor_pairing(10, 23)) # ça marche variment bien mais faire +1 pr eviter d'avoir 0 --> en fait je m'en fous car pourra jamais etre = 0
        print(_cantor_pairing(23, 10))
        print(_cantor_pairing(1, 32))
        print(_cantor_pairing(32, 1))

        # --> I love this because this is new and elegant


        # very good idea
        # I love this cantor formula
        sys.exit(0)


    # Open image and make into Numpy array
    # PILim = Image.open('patches.png').convert('RGB')
    RGBim = Img('/E/Sample_images/segmentation_assistant/ovipo_uncropped/200709_armGFP_suz_46hAPF_ON.lif - Series011/tracked_cells_resized.tif')[0:64, 0:64]
    # RGBim = np.array(PILim)

    # Make a single channel 24-bit image rather than 3 channels of 8-bit each

    print(RGBim.shape)

    # there is a bug but almost there still
    RGB24 = (RGBim[..., 0].astype(np.uint32) << 16) | (RGBim[..., 1].astype(np.uint32) << 8) | RGBim[..., 2].astype(
        np.uint32)

    print(RGB24.max(), RGB24.min())
    # RGB24 = (RGBim[...,0].astype(np.uint32)<<16) | (RGBim[...,1].astype(np.uint32)<<8) | RGBim[...,2].astype(np.uint32)

    # Run generic filter counting unique colours in neighbourhood
    result = detect_vertices_and_bonds(RGB24)


    # then could recolor them if needed --> could try that
    # give a unique id to bonds --> very easy in facts --> maybe replace old code ??? --> pb is that I need a uint64 image that will be quite big in mem but ok I guess or need relabel the image afterwards --> maybe by the way...
    result = generic_filter(RGB24.astype(np.uint64), partial(_identify_vertices_and_bonds,detect_bonds=True), (3, 3))
    # try to see how I can do that better!!!


    # result = generic_filter(RGB24.astype(float), partial(_identify_vertices_and_bonds,detect_bonds=True), (3, 3))
    print(result.dtype)
    # result = result.astype(np.uint8)

    # Save result
    # Image.fromarray(result.astype(np.uint8)).save('result.png')
    plt.imshow(result)
    plt.show()

    # plt.imshow(RGB24)
    # plt.show()

    indices = (result == 255)
    print('indices', indices)

    vertices = np.where(result == 255) #c'est
    print(vertices)
    if vertices[0].size != 0:
        print('y', vertices[0][0])
    if vertices[1].size != 0:
        print('x', vertices[1][0])
    if vertices[0].size + vertices[1].size != 0:
        print('first vx', result[vertices[0][0], vertices[1][0]])

    print(associate_vertices_to_cells(RGB24,
                                      result))  # ça marche en fait --> j'ai associé les vertex à chaque cellule --> cool
    # pb need add vertices at the corners too --> like the first or second pixel away from the border



from epyseg.deeplearning.deepl import EZDeepLearning
import os

def run_segmentation(deepTA=None, IS_TA_OUTPUT_MODE=True, input_channel_of_interest=None, TILE_WIDTH=256,
                     TILE_HEIGHT=256, TILE_OVERLAP=32, INPUT_FOLDER=None, EPYSEG_PRETRAINING='Linknet-vgg16-sigmoid-v2',
                     SIZE_FILTER=None, progress_callback=None):
    """
    Run the segmentation process.

    Args:
        deepTA (EZDeepLearning): An instance of EZDeepLearning class. If None, a new instance will be created.
        IS_TA_OUTPUT_MODE (bool): Indicates if the output should be in TA mode.
        input_channel_of_interest (int): The channel of interest in case of multichannel input image.
        TILE_WIDTH (int): Width of the input tiles for prediction.
        TILE_HEIGHT (int): Height of the input tiles for prediction.
        TILE_OVERLAP (int): Overlap between tiles for prediction.
        INPUT_FOLDER (str): Path to the folder containing the files to segment.
        EPYSEG_PRETRAINING (str): Pretraining model to use for segmentation.
        SIZE_FILTER (int): Size filter to remove cells below a certain pixel area.
        progress_callback (callable): Callback function for reporting progress.

    """
    if deepTA is None:
        deepTA = EZDeepLearning()
    deepTA.load_or_build(architecture='Linknet', backbone='vgg16', activation='sigmoid', classes=7,
                         pretraining=EPYSEG_PRETRAINING)

    deepTA.get_loaded_model_params()
    deepTA.summary()

    input_shape = deepTA.get_inputs_shape()
    output_shape = deepTA.get_outputs_shape()

    input_normalization = {'method': 'Rescaling (min-max normalization)', 'range': [0, 1], 'individual_channels': True}

    predict_parameters = {}
    predict_parameters["input_channel_of_interest"] = input_channel_of_interest
    predict_parameters["default_input_tile_width"] = TILE_WIDTH
    predict_parameters["default_input_tile_height"] = TILE_HEIGHT
    predict_parameters["default_output_tile_width"] = TILE_WIDTH
    predict_parameters["default_output_tile_height"] = TILE_HEIGHT
    predict_parameters["tile_width_overlap"] = TILE_OVERLAP
    predict_parameters["tile_height_overlap"] = TILE_OVERLAP
    predict_parameters["hq_pred_options"] = "Use all augs (pixel preserving + deteriorating) (Recommended for segmentation)"
    predict_parameters["post_process_algorithm"] = "default (slow/robust) (epyseg pre-trained model only!)"
    predict_parameters["input_normalization"] = input_normalization
    predict_parameters["filter"] = SIZE_FILTER

    predict_generator = deepTA.get_predict_generator(
        inputs=[INPUT_FOLDER], input_shape=input_shape, output_shape=output_shape, clip_by_frequency=None,
        **predict_parameters)

    if not IS_TA_OUTPUT_MODE:
        predict_output_folder = os.path.join(INPUT_FOLDER, 'predict')
    else:
        predict_output_folder = 'TA_mode'

    deepTA.predict(predict_generator, output_shape, predict_output_folder=predict_output_folder, batch_size=1,
                   progress_callback=progress_callback, **predict_parameters)

if __name__ == '__main__':
    INPUT_FOLDER = '/path/to/files_to_segment/'
    IS_TA_OUTPUT_MODE = True
    input_channel_of_interest = None
    TILE_WIDTH = 256
    TILE_HEIGHT = 256
    TILE_OVERLAP = 32
    EPYSEG_PRETRAINING = 'Linknet-vgg16-sigmoid-v2'
    SIZE_FILTER = None

    run_segmentation(INPUT_FOLDER=INPUT_FOLDER, IS_TA_OUTPUT_MODE=IS_TA_OUTPUT_MODE,
                     input_channel_of_interest=input_channel_of_interest, TILE_WIDTH=TILE_WIDTH,
                     TILE_HEIGHT=TILE_HEIGHT, TILE_OVERLAP=TILE_OVERLAP, EPYSEG_PRETRAINING=EPYSEG_PRETRAINING,
                     SIZE_FILTER=SIZE_FILTER)



import math

def compute_polarity_nematic(centroid, boundary_coords, image_to_compute_polarity_for):
    """
    Computes the polarity nematic tensor.

    Args:
        centroid (tuple): The centroid coordinates (y, x).
        boundary_coords (list): List of boundary coordinates.
        image_to_compute_polarity_for (numpy.ndarray): Image for polarity computation.

    Returns:
        list: The computed polarity values [S1, S2, total_intensity].

    """

    center_x = centroid[1]
    center_y = centroid[0]

    # Compute theta values for each boundary point
    theta_unsorted = []
    for pt in boundary_coords:
        deltaX = pt[1] - center_x
        deltaY = pt[0] - center_y
        rSquared = deltaX * deltaX + deltaY * deltaY
        r = math.sqrt(rSquared)
        cosTheta = deltaX / r
        sinTheta = deltaY / r
        theta = math.atan2(sinTheta, cosTheta)
        theta_unsorted.append(theta)

    # Create a hashmap with boundary points and corresponding theta values
    hashmap_point_angle = {}
    for z in range(len(boundary_coords)):
        hashmap_point_angle[z] = theta_unsorted[z]

    # Sort the hashmap by theta values
    cells_sorted_by_theta = {k: v for k, v in sorted(hashmap_point_angle.items(), key=lambda item: item[1])}

    # Compute necessary trigonometric quantities for each point
    theta_sorted = []
    cos2Theta_sorted = []
    sin2Theta_sorted = []
    for k, v in cells_sorted_by_theta.items():
        pt = boundary_coords[k]
        deltaX = pt[1] - center_x
        deltaY = pt[0] - center_y
        rSquared = deltaX * deltaX + deltaY * deltaY
        r = math.sqrt(rSquared)
        cosTheta = deltaX / r
        sinTheta = deltaY / r
        theta = math.atan2(sinTheta, cosTheta)
        theta_sorted.append(theta)
        cos2Theta_sorted.append(2. * cosTheta * cosTheta - 1.)
        sin2Theta_sorted.append(2. * cosTheta * sinTheta)

    # Compute polarity using the sorted values
    polarity = computeS1S2ForChannel(boundary_coords, cells_sorted_by_theta, theta_sorted, cos2Theta_sorted,
                                     sin2Theta_sorted, image_to_compute_polarity_for)

    return polarity


def computeS1S2ForChannel(boundary_points, cells_sorted_by_theta, theta_sorted, cos2Theta_sorted, sin2Theta_sorted,
                          channel_for_quantification):
    """
    Computes S1 and S2 values for a given channel.

    Args:
        boundary_points (list): List of boundary points.
        cells_sorted_by_theta (dict): Sorted boundary cells by theta.
        theta_sorted (list): Sorted theta values.
        cos2Theta_sorted (list): Sorted cos(2 * theta) values.
        sin2Theta_sorted (list): Sorted sin(2 * theta) values.
        channel_for_quantification (numpy.ndarray): Channel image for quantification.

    Returns:
        list: The computed S1, S2, and total intensity values.

    """

    k = 0
    S1_ch1 = 0
    S2_ch1 = 0
    sint_ch1 = 0.

    for idx in cells_sorted_by_theta.keys():
        pt = boundary_points[idx]
        curIntensity_ch1 = channel_for_quantification[pt[0], pt[1]]
        numPoints = len(boundary_points)

        # Compute theta values for neighboring points
        theta_last = theta_sorted[numPoints - 1] - (math.pi * 2.) if k == 0 else theta_sorted[k - 1]
        theta_next = theta_sorted[0] + (math.pi * 2.) if k == numPoints - 1 else theta_sorted[k + 1]

        # Compute deltaTheta as a three-point differential
        deltaTheta = 0.5 * (theta_next - theta_last)

        # Update S1, S2, and total intensity sums
        S1_ch1 += cos2Theta_sorted[k] * deltaTheta * curIntensity_ch1
        S2_ch1 += sin2Theta_sorted[k] * deltaTheta * curIntensity_ch1
        sint_ch1 += deltaTheta * curIntensity_ch1

        k += 1

    return [S1_ch1.tolist(), S2_ch1.tolist(),
            sint_ch1.tolist()]



import os
from epyseg.settings.global_settings import set_UI
set_UI()
from qtpy.QtCore import QTimer


class Blinker:
    """
    Utility class for blinking components by changing their background color to red temporarily.

    Methods:
        blink(component): Blinks the given component or a list of components.
        reset_stylesheet(component): Resets the stylesheet of the given component or a list of components.

    """

    def __init__(self):
        pass

    def blink(self, component):
        """
        Blinks the given component or a list of components by changing their background color to red temporarily.

        Args:
            component (QWidget or list[QWidget]): The component or list of components to be blinked.

        """
        if isinstance(component, list):
            for comp in component:
                comp.setStyleSheet("background-color: red;")
        else:
            component.setStyleSheet("background-color: red;")

        self.current_timer = QTimer()
        self.current_timer.timeout.connect(lambda: self.reset_stylesheet(component))
        self.current_timer.setSingleShot(True)
        self.current_timer.start(3000)

    def reset_stylesheet(self, component):
        """
        Resets the stylesheet of the given component or a list of components, restoring the original appearance.

        Args:
            component (QWidget or list[QWidget]): The component or list of components to reset the stylesheet.

        """
        if isinstance(component, list):
            for comp in component:
                comp.setStyleSheet('')
        else:
            component.setStyleSheet('')


if __name__ == '__main__':
    blinker = Blinker()
    # blinker.blink()



# modified from https://stackoverflow.com/questions/24469662/how-to-redirect-logger-output-into-pyqt-text-widget
import os
from epyseg.settings.global_settings import set_UI
set_UI()
import sys
from qtpy import QtCore
import logging


class QtHandler(logging.Handler):
    """
    Custom logging handler for Qt application.
    It emits log records to either stdout or stderr based on the log level.

    Attributes:
        messageWritten (QtCore.Signal): Signal emitted when a log message is written.

    Methods:
        emit(record): Formats and emits the log record.

    """

    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record):
        """
        Formats and emits the log record.

        Args:
            record (logging.LogRecord): The log record to be emitted.

        """
        record = self.format(record)
        if record:
            if record.startswith('ERROR') or record.startswith('CRITICAL') or record.startswith('WARNING'):
                XStream.stderr().write('%s' % record)
            else:
                XStream.stdout().write('%s' % record)


class XStream(QtCore.QObject):
    """
    Custom QObject class for capturing stdout and stderr.

    Attributes:
        _stdout (XStream): The singleton instance for stdout.
        _stderr (XStream): The singleton instance for stderr.
        messageWritten (QtCore.Signal): Signal emitted when a log message is written.

    Methods:
        flush(): Flushes the output.
        fileno(): Returns the file number.
        write(msg): Writes the message to the output stream.
        stdout(): Returns the singleton instance for stdout.
        stderr(): Returns the singleton instance for stderr.

    """

    _stdout = None
    _stderr = None
    messageWritten = QtCore.Signal(str)

    def flush(self):
        """
        Flushes the output.

        """
        pass

    def fileno(self):
        """
        Returns the file number.

        Returns:
            int: The file number (-1).

        """
        return -1

    def write(self, msg):
        """
        Writes the message to the output stream.

        Args:
            msg (str): The message to be written.

        """
        if not self.signalsBlocked():
            self.messageWritten.emit(msg)

    @staticmethod
    def stdout():
        """
        Returns the singleton instance for stdout.

        Returns:
            XStream: The singleton instance for stdout.

        """
        if not XStream._stdout:
            XStream._stdout = XStream()
            sys.stdout = XStream._stdout
        return XStream._stdout

    @staticmethod
    def stderr():
        """
        Returns the singleton instance for stderr.

        Returns:
            XStream: The singleton instance for stderr.

        """
        if not XStream._stderr:
            XStream._stderr = XStream()
            sys.stderr = XStream._stderr
        return XStream._stderr



import random
import matplotlib.pyplot as plt
import numpy as np

def plot_image_with_pixels(image, cmap='gray'):
    '''
    Plots an image with white lines surrounding each pixel. NB: This only make sense for small images

    Args:
        image (numpy.ndarray): The input image array.
        cmap (str, optional): The colormap to use for displaying the image. Defaults to 'gray'.

    Returns:
        matplotlib.figure.Figure: The figure object containing the plotted image.

    # Examples:
    #     >>> image = np.random.rand(10, 10)
    #     >>> fig = plot_image_with_pixels(image)
    #     >>> plt.show()
    '''
    # Create a new figure and axes
    fig, ax = plt.subplots()

    # Show the image using imshow()
    ax.imshow(image, cmap=cmap, interpolation=None)

    # Set the tick locations to center the ticks at the pixels
    ax.set_xticks(np.arange(image.shape[1]), minor=False)
    ax.set_yticks(np.arange(image.shape[0]), minor=False)

    # Shift the grid by 0.5 to surround the pixels
    ax.set_xticks(np.arange(image.shape[1]+1)-0.5, minor=True)
    ax.set_yticks(np.arange(image.shape[0]+1)-0.5, minor=True)

    # If the image is too large, hide the tick labels
    if image.shape[0] > 50:
        ax.set_yticklabels([])

    if image.shape[1] > 50:
        ax.set_xticklabels([])

    # Add the lines around each pixel using grid()
    ax.grid(which='minor', color='white', linestyle='-', linewidth=1)

    # Return the figure object
    return fig


if __name__ == '__main__':

    if True:
        # Generate a random image of a random type
        image = random.choice([np.random.rand(10, 10), np.random.rand(100, 10, 3)])

        # Plot the image with white lines surrounding each pixel
        fig = plot_image_with_pixels(image)

        # Show the plot
        plt.show()


    if True:
        from epyseg.img import pop, Img
        img = Img('/E/Sample_images/sample_images_FIJI/AuPbSn40.jpg')
        pop(img)
        # sys.exit(0)







import numpy as np

# Create a 3D array with 100 rows
arr = np.random.rand(100, 3)

# Create a boolean index of length 100 that has only 5 True values
bool_index = np.zeros(100, dtype=bool)
bool_index[:5] = True
np.random.shuffle(bool_index)

print(bool_index.shape)
# Filter the rows of the 3D array using the boolean index
filtered_arr = arr[bool_index]

print(filtered_arr.shape)